'use strict';

var stream = require('readable-stream');

function asCallback (fn) {
  if (typeof fn !== 'function') {
    throw new TypeError('A callback must be a function')
  }
  fn[Symbol.for('callback')] = true;
  return fn
}

function isCallback (fn) {
  return typeof fn === 'function' && Symbol.for('callback') in fn
}

function once (callback) {
  let called = false;
  return function (err, res) {
    if (called) return
    called = true;
    callback(err, res);
  }
}

function makeConcurrentMethods (concurrency, _write, _final) {
  if (!Number.isInteger(concurrency) || concurrency < 1) {
    throw new TypeError('Concurrency must be a positive integer')
  }
  if (concurrency === 1) {
    return [_write, _final]
  }

  // Active jobs counter
  let jobs = 0;

  // First encountered error
  let reason = null;

  // Job end callback
  let done = noop;

  function concurrentWrite (chunk, encoding, callback) {
    if (reason) {
      callback(reason);
      return
    }

    jobs++;
    _write(chunk, encoding, (err, result) => {
      jobs--;

      if (err && !reason) {
        reason = err;
      }

      if (!reason && result !== undefined && this.push) {
        this.push(result);
      }

      done(reason);
    });

    if (jobs < concurrency) {
      callback(reason);
    } else {
      done = once(callback);
    }
  }

  function concurrentFinal (callback) {
    done = () => {
      if (reason) {
        callback(reason);
      } else if (jobs <= 0) {
        if (_final) {
          _final(callback);
        } else {
          callback(null);
        }
      }
    };

    if (jobs <= 0) {
      done();
    }
  }

  return [concurrentWrite, concurrentFinal]
}

function noop () {
  // nothing to do
}

function patchWritable (writable, concurrency) {
  const [_write, _final] = makeConcurrentMethods(
    concurrency,
    bind(writable._write, writable),
    bind(writable._final, writable)
  );
  writable._write = _write;
  writable._final = _final;
}

function patchTransform (transform, concurrency) {
  const [_transform, _flush] = makeConcurrentMethods(
    concurrency,
    bind(transform._transform, transform),
    bind(transform._flush, transform)
  );
  transform._transform = _transform;
  transform._flush = _flush;
}

function bind (fn, obj) {
  return typeof fn === 'function'
    ? fn.bind(obj)
    : fn
}

function handlePromise (fn) {
  if (typeof fn !== 'function') {
    return fn
  }
  return Object.defineProperty(
    function (...args) {
      const callback = args[args.length - 1];
      const value = fn.call(this, ...args);
      if (isPromiseLike(value)) {
        value.then(
          result => callback(null, result),
          callback
        );
      }
    },
    'name',
    { value: fn.name }
  )
}

function isPromiseLike (value) {
  return typeof Object(value).then === 'function'
}

function createReadMethod (asyncRead) {
  let reading = false;

  return function _read (size) {
    if (reading) {
      return
    }
    reading = true;

    const callback = once((err, data) => {
      reading = false;

      // Detects manual .push(null) and .destroy()
      const alive = !this.closed && !this.readableEnded;
      if (!alive) {
        return
      }

      if (err) {
        // This will set `closed` to true
        return this.destroy(err)
      }

      if (data !== undefined) {
        // If null, this will set `readableEnded` to true
        this.push(data);
      }

      if (
        data !== null &&
        this.readableFlowing &&
        this.readableLength < this.readableHighWaterMark
      ) {
        // We have room inside the buffer
        process.nextTick(() => this._read(size));
      }
    });

    const out = asyncRead.call(this, size, callback);
    if (isPromiseLike(out)) {
      out.then(
        result => callback(null, result),
        callback
      );
    }
  }
}

class Duplex extends stream.Duplex {
  constructor (options) {
    super(options);

    if (typeof Object(options).asyncRead === 'function') {
      this._asyncRead = options.asyncRead;
    }

    this._write = handlePromise(this._write);
    this._writev = handlePromise(this._writev);
    this._destroy = handlePromise(this._destroy);
    this._final = handlePromise(this._final);
    this._construct = handlePromise(this._construct);

    if (this._asyncRead) {
      this._read = createReadMethod(this._asyncRead);
    }

    const concurrency = Object(options).concurrency;
    if (Number.isInteger(concurrency) && concurrency >= 2) {
      patchWritable(this, concurrency);
    }
  }
}

function finished (...args) {
  return typeof args[args.length - 1] === 'function'
    ? stream.finished(...args)
    : stream.promises.finished(...args)
}

function isNodeStream (obj) {
  return isReadableStream(obj) || isWritableStream(obj)
}

function isReadableStream (value) {
  return typeof value === 'object' &&
    value !== null &&
    typeof value.on === 'function' &&
    typeof value.pipe === 'function' &&
    typeof value.push === 'function'
}

function isWritableStream (value) {
  return typeof value === 'object' &&
    value !== null &&
    typeof value.on === 'function' &&
    typeof value.write === 'function' &&
    typeof value.end === 'function'
}

function isDuplexStream (value) {
  return isReadableStream(value) && isWritableStream(value)
}

function pipeline (...args) {
  return isCallback(args[args.length - 1])
    ? stream.pipeline(...args)
    : stream.promises.pipeline(...args)
}

class Readable extends stream.Readable {
  constructor (options) {
    super(options);

    if (typeof Object(options).asyncRead === 'function') {
      this._asyncRead = options.asyncRead;
    }

    this._destroy = handlePromise(this._destroy);
    this._construct = handlePromise(this._construct);

    if (this._asyncRead) {
      if (this._read !== Readable.prototype._read) {
        throw new Error('The _read method cannot be defined along with the _asyncRead method')
      }
      this._read = createReadMethod(this._asyncRead);
    }
  }
}

class Transform extends stream.Transform {
  constructor (options) {
    super(options);

    this._destroy = handlePromise(this._destroy);
    this._construct = handlePromise(this._construct);
    this._transform = handlePromise(this._transform);
    this._flush = handlePromise(this._flush);

    const concurrency = Object(options).concurrency;
    if (Number.isInteger(concurrency) && concurrency >= 2) {
      patchTransform(this, concurrency);
    }
  }
}

class Writable extends stream.Writable {
  constructor (options) {
    super(options);

    this._write = handlePromise(this._write);
    this._writev = handlePromise(this._writev);
    this._destroy = handlePromise(this._destroy);
    this._final = handlePromise(this._final);
    this._construct = handlePromise(this._construct);

    const concurrency = Object(options).concurrency;
    if (Number.isInteger(concurrency) && concurrency >= 2) {
      patchWritable(this, concurrency);
    }
  }
}

Object.defineProperty(exports, 'PassThrough', {
  enumerable: true,
  get: function () { return stream.PassThrough; }
});
Object.defineProperty(exports, 'Stream', {
  enumerable: true,
  get: function () { return stream.Stream; }
});
Object.defineProperty(exports, 'addAbortSignal', {
  enumerable: true,
  get: function () { return stream.addAbortSignal; }
});
Object.defineProperty(exports, 'compose', {
  enumerable: true,
  get: function () { return stream.compose; }
});
Object.defineProperty(exports, 'isErrored', {
  enumerable: true,
  get: function () { return stream.isErrored; }
});
Object.defineProperty(exports, 'isReadable', {
  enumerable: true,
  get: function () { return stream.isReadable; }
});
exports.Duplex = Duplex;
exports.Readable = Readable;
exports.Transform = Transform;
exports.Writable = Writable;
exports.asCallback = asCallback;
exports.finished = finished;
exports.isCallback = isCallback;
exports.isDuplexStream = isDuplexStream;
exports.isNodeStream = isNodeStream;
exports.isReadableStream = isReadableStream;
exports.isWritableStream = isWritableStream;
exports.pipeline = pipeline;
